#include <sys/epoll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#ifndef OPEN_MAX
#define OPEN_MAX 1024
#endif
#define SERV_PORT 8888
#define MAXLINE 4096
#define EVENT_MAX 20
struct event_data{
	int fd;
	int offset;
	char buf[MAXLINE];
}event_d[OPEN_MAX];
void set_event_d(int fd, struct epoll_event *evt, struct event_data *met){
	met->fd = fd;
	met->offset = 0;
	memset(&met->buf, 0, sizeof met->buf);
	evt->data.ptr = met;
}
int main(int , char *[]){
	int listenfd, connfd, epfd;
//    char buf[MAXLINE];
    struct sockaddr_in servaddr, cliaddr;
    int i, j, nready;
	socklen_t clilen;
    ssize_t n, wpos;
    struct epoll_event evt, evts[EVENT_MAX];
    listenfd = socket(AF_INET, SOCK_STREAM, 0);
	//设置lisenfd非阻塞
	fcntl(listenfd, F_SETFL, O_NONBLOCK);
    memset(&servaddr, 0, sizeof servaddr);
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SERV_PORT);
    bind(listenfd, (struct sockaddr*)&servaddr, sizeof servaddr);
    listen(listenfd, 1024);
	//创建epoll描述符
    epfd = epoll_create(OPEN_MAX);
	
	//将listen描述符加入到epoll中
	set_event_d(listenfd, &evt, &event_d[0]);
	evt.events = EPOLLIN | EPOLLET;
    epoll_ctl(epfd, EPOLL_CTL_ADD, listenfd, &evt);
	
	//event_d中的fd为-1表示没有使用，0用来存listenfd，其它用来存客户端fd
	for (i = 1; i < OPEN_MAX; i++) event_d[i].fd = -1;
    for (;;) {
	    nready = epoll_wait(epfd, evts, EVENT_MAX, -1);
	    for (i = 0; i < nready; i++) {
			//读取存储描述符信息的指针
			struct event_data *ed = (struct event_data*)evts[i].data.ptr;
			
			//accept
            if (ed->fd == listenfd) {
            	//ET模式下存在多个client connect只通知一次的情况,需要循环accept直到读到EAGAIN
				for(;;) {
					clilen = sizeof cliaddr;
					connfd = accept(listenfd, (struct sockaddr*)&cliaddr, &clilen);
					if (connfd == -1) {
						if (errno == EAGAIN) break;
						else perror("accept error");
					} else {
						printf("a client connected! fd: %d\n", connfd);
					}
					//找到可用的event_d来存放event.data
					for (j = 1; j < OPEN_MAX; j++) {
						if (event_d[j].fd == -1) break;
					}
					if (j == OPEN_MAX) {
						perror("too many clients");
						break;
					}
					//设置客户端fd非阻塞
					fcntl(connfd, F_SETFL, O_NONBLOCK);
					set_event_d(connfd, &evt, &event_d[j]);
					evt.events = EPOLLIN | EPOLLET;
					epoll_ctl(epfd, EPOLL_CTL_ADD, connfd, &evt);
				}
			//read
			} else if(evts[i].events & EPOLLIN){
				//ET模式，重复读直到EAGAIN说明无数据可读或者读到错误及EOF
				for (;ed->offset < MAXLINE;) {
					n = read(ed->fd, ed->buf + ed->offset, MAXLINE - ed->offset);
					
					if (n <= 0) {
						if (errno == EAGAIN) break;
						if (errno == EINTR) continue;
						close(ed->fd);
						ed->fd = -1;
						break;
					}
					ed->offset += n;
				}
				//修改为监听描述符写就绪
				evt.events = EPOLLOUT | EPOLLET;
				evt.data.ptr = ed;
				epoll_ctl(epfd, EPOLL_CTL_MOD, ed->fd, &evt);
			//write
			} else if(evts[i].events & EPOLLOUT){
				wpos = 0;
				//ET模式下，重复写直到无数据可发或者EAGAIN
				for (;wpos < ed->offset;) {
					n = write(ed->fd, ed->buf + wpos, ed->offset - wpos);
					if (n < 0) {
						if (errno == EAGAIN) break;
						if (errno == EINTR) continue;
						close(ed->fd);
						ed->fd = -1;
						break;
					}
					wpos += n;
				}
				ed->offset = 0;
				//修改为监听描述符读就绪
				evt.events = EPOLLIN | EPOLLET;
				evt.data.ptr = ed;
				epoll_ctl(epfd, EPOLL_CTL_MOD, ed->fd, &evt);
			}
        }
    }
	return 0;
}
